const { checkRule, hideStack } = require('./lib');


/**
 * @typedef {import('.').Rule} Rule
 */


/**
  * SyncReplaceable function receives the whole string and returns the result of transforms which are of interface `{ re: /reg-(exp)/, replacement(m, t) { return `transform-${t}` } }`
  * If the replacement is passed as a function, it will work in the same way as the replacer for `string.replace` method (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace), taking the `match` as the first argument, and matched `p1`, `p2`, _etc_ parameters as following arguments. The replacer can also be an async function.
  * @param {string|Buffer} input The string or buffer to transform synchronously using the replacements. Does not support asynchronous replacements.
  * @param {Rule[]} rules An array with rules.
  * @example
  *
  * // markdown __ to html emphasise implementation
  * const stream = SyncReplaceable('__hello world__', {
  *  --re: /__(\S+)__/g,
  *  --replacement(match, p1) {
  *  ----return `<em>${p1}</em>`
  *  --},
  * })
  */
const SyncReplaceable = (input, rules) => {
  const fr = rules.filter(checkRule)
  const s = fr.reduce((acc, { re, replacement }) => {
    /** @type {string} */
    let Acc = acc
    if (this._broke) return Acc

    if (typeof replacement == 'string') {
      Acc = Acc.replace(re, replacement)
    } else {
      /** @type {function} */
      const R = replacement.bind(this)
      let commonError
      const t = Acc.replace(re, (match, ...args) => {
        commonError = new Error()
        try {
          if (this._broke) return match
          const p = R(match, ...args)
          return p
        } catch (e) { // hide stack for sync stack traces
          hideStack(commonError, e)
        }
      })
      return t
    }
  }, `${input}`)
  return s
}

module.exports=SyncReplaceable

//   constructor(rules) {
//     const re = Array.isArray(rules) ? rules : [rules]
//     const fr = re.filter(checkRule)
//     this.rules = fr
//   }

//   /**
//    * Stop executing further after the current rule.
//    */
//   brake() {
//     this._broke = true
//   }

//   async reduce(chunk) {
//     /** @type {string} */
//     const s = await this.rules.reduce(async (acc, { re, replacement }) => {
//       /** @type {string} */
//       let string = await acc
//       if (this._broke) return string

//       if (typeof replacement == 'string') {
//         string = string.replace(re, replacement)
//       } else {
//         /** @type {function} */
//         const R = replacement.bind(this)
//         const promises = []
//         let commonError
//         const t = string.replace(re, (match, ...args) => {
//           commonError = new Error()
//           try {
//             if (this._broke) return match
//             const p = R(match, ...args)
//             if (p instanceof Promise) {
//               promises.push(p)
//             }
//             return p
//           } catch (e) { // hide stack for sync stack traces
//             hideStack(commonError, e)
//           }
//         })
//         if (promises.length) {
//           try { // hide stack only for when throw happens before awaits
//             const data = await Promise.all(promises)
//             string = string.replace(re, () => data.shift())
//           } catch (e) {
//             hideStack(commonError, e)
//           }
//         } else {
//           string = t
//         }
//       }
//       return string
//     }, `${chunk}`)

//     return s
//   }
//   async _transform(chunk, _, next) {
//     try {
//       const s = await this.reduce(chunk)
//       this.push(s)
//       next()
//     } catch (e) {
//       const s = cleanStack(e.stack)
//       e.stack = s
//       next(e)
//     }
//   }
// }