const getDefaultRegExp = (name) => {
  return new RegExp(`%%_RESTREAM_${name.toUpperCase()}_REPLACEMENT_(\\d+)_%%`, 'g')
}

/**
 * Make a unique string to replace cut text with.
 * @param {string} name
 * @param {number} index
 */
const getDefaultReplacement = (name, index) => {
  return `%%_RESTREAM_${name.toUpperCase()}_REPLACEMENT_${index}_%%`
}

/**
 * Create a new marker.
 * @param {string} name The name of the matcher, used in the doc marker.
 * @param {RegExp} re The regular expression used for detection.
 * @return {Marker} The marker.
 */
const makeMarker = (name, re, {
  getReplacement = getDefaultReplacement,
  getRegex = getDefaultRegExp,
} = {}) => {
  const regExp = getRegex(name)
  return {
    name,
    re,
    regExp,
    getReplacement,
    map: {},
    lastIndex: 0,
  }
}

/**
 * Make markers from a configuration object.
 * @param {Object.<string, RegExp>} matchers An object with types of markers to create as keys and their detection regexes as values.
 * @param {MakeMarkersConfig} [config] Additional configuration.
 * @param {(name: string, index: number) => string} [config.getReplacement] A function used to create a replacement when some text needs to be cut.
 * @param {(name: string) => RegExp} [config.getRegex] A function used to create a RegExp to detect replaced chunks.
 * @returns {Object.<string, Marker>} An object with markers for each requested type.
 */
       const makeMarkers = (matchers, config) => {
  const res = Object.keys(matchers).reduce((acc, key) => {
    const re = matchers[key]
    const marker = makeMarker(key, re, config)
    const m = {
      ...acc,
      [key]: marker,
    }
    return m
  }, {})
  return res
}

/**
 * Make a rule for pasting markers back.
 * @param {Marker} marker A marker used to cut and paste portions of text to exclude them from processing by other rules.
 * @returns {Rule} A rule to paste previously replaced chunks.
 */
       const makePasteRule = (marker) => {
  const { regExp: re, map } = marker
  const rule = {
    re,
    replacement(match, index) {
      const m = map[index]
      delete map[index]
      return m
    },
  }
  return rule
}

/**
 * Make a rule for initial replacement of markers.
 * @param {Marker} marker A marker used to cut and paste portions of text to exclude them from processing by other rules.
 * @returns {Rule} A rule to cut matched chunks.
 */
       const makeCutRule = (marker) => {
  const { re, map, getReplacement, name } = marker
  const rule = {
    re,
    replacement(match) {
      const { lastIndex } = marker
      map[lastIndex] = match
      marker.lastIndex += 1
      const m = getReplacement(name, lastIndex)
      return m
    },
  }
  return rule
}

/**
 * @typedef {import('..').Rule} Rule
 */

/* documentary types/markers.xml */
/**
 * @typedef {Object} Marker A marker used to cut and paste portions of text to exclude them from processing by other rules.
 * @prop {string} type Type of the marker.
 * @prop {(index) => string} getMarker A function to generate markers which can be then found.
 * @prop {RegExp} re A regular expression used for detection of the match.
 * @prop {RegExp} regExp A generated regular expression to replace the marker back to its original value.
 * @prop {object} map A map which holds detected matches at their indexes.
 * @prop {number} lastIndex An index of the last inserted element. Starts with 0.
 *
 * @typedef {Object} MakeMarkersConfig Additional configuration.
 * @prop {(name: string, index: number) => string} [getReplacement] A function used to create a replacement when some text needs to be cut.
 * @prop {(name: string) => RegExp} [getRegex] A function used to create a RegExp to detect replaced chunks.
 */


module.exports.makeMarkers = makeMarkers
module.exports.makePasteRule = makePasteRule
module.exports.makeCutRule = makeCutRule